(in-package :krb-util)

(use-package :incf-cl)

(defgeneric ->string (thing)
  (:documentation "Convert a something into a string."))

(defmethod ->string ((thing null))
  "")

;; is this type 'string' type?  it is according to type-of in the
;; slime repl for SBCL
(defmethod ->string ((thing simple-array))
  thing)

(defmethod ->string ((thing t))
  "Convert things to a string representation.  This is just short hand
for format (the author was used to it from Scheme)."
  (format nil "~a" thing))

;; I should make these (lc/uc/->lc/->uc and even ->string) inline...
(defun lc (thing)
  (format nil "~(~a~)" thing))

(defun uc (thing)
  (format nil "~:@(~a~)" thing))

(defun ->lc (thing)
  (lc (->string thing)))

(defun ->uc (thing)
  (uc (->string thing)))


(defvar *newline* (->string #\Newline))
(defvar *lf*      (->string (code-char 10)))
(defvar *cr*      (->string (code-char 13)))
(defvar *crlf*    (format nil "~a~a" *cr* *lf*))
(defvar *tab*     (->string #\Tab))
(defvar *space*   (->string #\Space))
(defvar *comma*   (->string ","))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Anaphoric macros (from _On Lisp_ by Paul Graham)

(defmacro aif (if then &optional else)
  "Anaphoric if - binds the symbol 'it to the result of the condition."
  `(let ((it ,if))
     (if it
       ,then ,else)))

(defmacro awhen (test &body body)
  "Anaphoric when - binds the symbol 'it to the result of the condition."
  `(aif ,test
        (progn ,@body)))

(defmacro aand (&rest args)
  "Anaphoric and - binds the symbol 'it to the result of the most recently evaluated condition."
  (cond ((null args) t)
        ((null (cdr args)) (car args))
        (t `(aif ,(car args) (aand ,@(cdr args))))))

(defmacro acond (&rest clauses)
  "Anaphoric cond - binds the symbol 'it to the condition result."
  (when clauses
    (let ((clause (car clauses))
          (sym (gensym)))               ; used so that it is bound only around rhs of clause
      `(let ((,sym ,(car clause)))
         (if ,sym
           (let ((it ,sym)) ,@(cdr clause))
           (acond ,@(cdr clauses)))))))

(defmacro alambda (args &body body)
  "Anaphoric lambda - binds the symbol 'self to the lambda itself."
  `(labels ((self ,args ,@body))
     #'self))

(defmacro aprog1 (init &body body)
  "Anaphoric prog1 - binds the symbol 'it to the result of evaluating
the first form, subsequent forms are evaluated for effect, just as
with prog1.

My favorite use for this is in-place construction of hash-tables:

  (aprog1
    (make-hash-table :test #'equal)
    (setf (gethash it :a) 1)
    (setf (gethash it :b) 2)
    (setf (gethash it :c) 3))

"
  `(let ((it ,init))
     ,@body
     it))

(defmacro until (test &body body)
  `(do ()
       (,test)
     ,@body))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric ->vector (thing))

(defgeneric ->list (thing))

(defmethod ->vector ((lst list))
  (aprog1 (make-array (length lst))
    (loop for item in lst
       for ii = 0 then (1+ ii)
       do
	 (setf (aref it ii)
	       item))))

(defmethod ->vector ((thing vector))
  thing)

(defmethod ->list ((vec vector))
  (loop for item across vec
       collect item))

(defmethod ->list ((thing list))
  thing)

(define-condition record-too-long (error)
  ((text :initarg :text :reader text)))

(defgeneric ensure-length (len things))

(defmethod ensure-length ((len integer)
			  (things list))
  (let ((nthings (length things)))
    (cond ((> nthings len)
	   (error 'record-too-long))
	  ((= nthings len)
	   things)
	  (t
	   (loop
	      for ii from 0 upto (1- len)
	      collect (if (>= ii nthings)
			  nil
			  (nth ii things)))))))

(defmethod ensure-length ((len integer)
			  (things simple-vector))
  (let ((nthings (length things)))
    (cond ((> nthings len)
	   (error 'record-too-long))
	  ((= nthings len)
	   things)
	  (t
	   (->vector
	    (loop
	       for ii from 0 upto (1- len)
	       with len = (length things)
	       collect (if (>= ii len)
			   nil
			   (aref things ii))))))))



(defun hash-table-keys (ht)
  "Returns the keys of the given hashtable as a list."
  (loop for k being the hash-keys of ht
       collect k))

(defun copy-hash-table (ht)
  "Make an idendical copy of the given hashtable"
  (loop with nht = (make-hash-table :test (hash-table-test ht))
       for k being the hash-keys in ht using (hash-value v)
       finally (return nht)
       do
       (setf (gethash k nht) v)))

(defun plist-get (plist key)
  (cond ((null plist)
	 nil)
	((equal key (first plist))
	 (second plist))
	(t
	 (plist-get (cddr plist)
		     key))))

(defun plist->assoc (plist)
  (labels ((convert (acc plist)
	     (cond ((null plist)
		    acc)
		   (t
		    (convert (cons (cons (first plist)
					 (second plist))
				   acc)
			     (cddr plist))))))
    (convert (list)
	      plist)))


(define-condition missing-assoc-entry (error)
  ((text :initarg :text :reader text)))

(defun eassoc (key alist)
  (let ((pair (assoc key alist)))
    (when (null pair)
      (error 'missing-assoc-entry :text (format nil "missing entry:~a from assoc list:~a" key alist)))
    (cdr pair)))

(defun intersperse (delim things)
  (labels ((isp (things acc)
	     (cond ((null things)
		  acc)
		 (t
		  (isp (cdr things)
		       (cons (car things) (cons delim
					  acc)))))))
    (cond ((null things)
	   nil)
	  (t
	   (reverse (isp (cdr things)
		  (list (car things))))))))

(defun flatten (&rest trees)
  (cond ((null trees)
	 (list))
	((listp (car trees))
	 (append
	  (apply #'flatten (car trees))
	  (apply #'flatten (cdr trees))))
	(t
	 (cons (car trees)
	       (apply #'flatten (cdr trees))))))

(defun string-join (delim &rest things)
  (apply #'concatenate (cons 'string (mapcar #'->string (intersperse delim (flatten things))))))


(defun curry (fn &rest args)
  #'(lambda (&rest more)
      (apply fn (append args more))))

(defun rcurry (fn &rest args)
  #'(lambda (&rest more)
      (apply fn (append more args))))

(defun ssplit (delim string &optional limit)
  (cl-ppcre:split delim string :limit limit))


(defun all-permutations (things)
  (cond ((= 1 (length things))
	 (list things))
	(t
	 (assemble
	     (cons head tail)
	   (<- head things)
	   (<- tail (all-permutations (remove head things)))))))

;; (all-permutations '(a b c))


(defun symbol-equal (a b)
  (string= (format nil "~a" a)
	   (format nil "~a" b)))


(defmacro cut (fn &rest body)
  "
(cut #'cons (+ a 1) <>) is the same as (lambda (x2) (cons (+ a 1) x2))
(cut #'list 1 <> 3 <> 5) is the same as (lambda (x2 x4) (list 1 x2 3 x4 5))
(cut #'list) is the same as (lambda () (list))
(cut #'list 1 <> 3 <...>) is the same as (lambda (x2 . xs) (apply list 1 x2 3 xs))

The following form is not supported in this version:
  (cut <> a b) is the same as (lambda (f) (f a b))
"
  (let* ((formals (list))
	 (new-body 
	  (mapcar #'(lambda (item)
		      (if (symbol-equal '<> item)
			  (aprog1 
			   (gensym)
			   (push it formals))
			  item))
		  body)))
    `#'(lambda ,(reverse formals)
	 (funcall ,fn ,@new-body))))

(defun map-tree (fn tree)
  (cond ((null tree)
	 tree)
	((not (listp tree))
	 (funcall fn tree))
	(t
	 (mapcar (curry #'map-tree fn) tree))))

;; (macroexpand-1 '(curry #'map-tree fn))

;; (map-tree #'(lambda (elt) (format t "x:~a~&" elt) (1+ elt)) '(1 2 (3 4 (5 6 (7)))))

(defmacro rcut (&rest body)
  "'Recrusive' `cut'.  Like `cut', but you don't have to destrcuture
the arguments yourself, you can use any form you normally would with
the cut points factored out:

  (rcut (format nil \"thing is: ~a~&\" <>))

is equivalent to 

  (cut #'format nil \"thing is: ~a~&\" <>)

`rcut' can also be used with more complicated forms, it will
recursivly process the body form given to it - this is also a point of
caution since nesting forms which use the target symbol '<>' may not
have the effect you expect."
  (when (null body)
    (error "krb-util:rcut : Error, body can't be null in macro (rcut ...)"))
  (let* ((formals (list))
	 (new-body 
	  (map-tree #'(lambda (elt)
			(if (symbol-equal '<> elt)
			    (aprog1
			     (gensym)
			     (push it formals)
			     it)
			    elt))
		    body)))
    `#'(lambda ,(reverse formals)
	 ,@new-body)))


(defun starts-with-? (sym)
  (equal "?"  (subseq (format nil "~a" sym) 0 1)))


(defmacro pcut (&rest body)
  (when (null body)
    (error "krb-util:pcut : Error, body can't be null in macro (pcut ...)"))
  (let* ((formals (list))
         (fml-hash (make-hash-table :test #'equal))
         (new-body
          (map-tree #'(lambda (elt)
                        (if (starts-with-? elt)
                            (aprog1
                             (or (gethash elt fml-hash)
                                 (aprog1
                                  (gensym)
                                  (push it formals)
                                  (setf (gethash elt fml-hash) it)
                                  it))
                             it)
                            elt))
                    body)))
    `#'(lambda ,(reverse formals)
         ,@new-body)))

(defmacro kcut (&rest body)
  (when (null body)
    (error "krb-util:kcut : Error, body can't be null in macro (kcut ...)"))
  (let* ((formals (list))
         (fml-hash (make-hash-table :test #'equal))
         (new-body
          (map-tree #'(lambda (elt)
                        (if (starts-with-? elt)
                            (aprog1
                             (or (gethash elt fml-hash)
                                 (aprog1
                                  (read-from-string (subseq (format nil "~a" elt) 1))
                                  (push it formals)
                                  (setf (gethash elt fml-hash) it)))
                             it)
                            elt))
                    body)))
    `#'(lambda (&key ,@(reverse formals))
         ,@new-body)))

(defun list-uniq (things)
  (let ((ht (make-hash-table :test #'equal))
	(res (list)))
    (loop for thing in things
       do
	 (when (not (gethash thing ht))
	   (setf (gethash thing ht) t)
	   (push thing res)))
    (reverse res)))


(defun expand-tilde-file (path)
  (cond ((equal "~/" (subseq path 0 2))
         (format nil "~a/~a"
                 (sb-posix:getenv "HOME")
                 (subseq path 2)))
        ((equal "~" (subseq path 0 1))
         (error "krb-util:misc-util.lisp/expand-tilde-file: implement ~user-name expansion"))))


;(expand-tilde-file "~/foo")
